;; -*- no-byte-compile: t; lexical-binding: nil -*-
(define-package "cape" "20260519.1021"
  "Completion At Point Extensions."
  '((emacs  "29.1")
    (compat "31"))
  :url "https://github.com/minad/cape"
  :commit "c99911b08831c26179145686b4beffa96f1f8a68"
  :revdesc "c99911b08831"
  :keywords '("abbrev" "convenience" "matching" "completion" "text")
  :authors '(("Daniel Mendler" . "mail@daniel-mendler.de"))
  :maintainers '(("Daniel Mendler" . "mail@daniel-mendler.de")))
